//-----------------------------
// CSEE 4840 Embedded System Design - System Constraints File
//
// Webserver
// Franklin Ma, Howard Wang, Victor Wang, William Wong
//
//-----------------------------

#include "xio.h"
#include "xbasic_types.h"

#ifndef BYTE
#define BYTE unsigned char
#endif
#ifndef WORD
#define WORD unsigned short
#endif
#ifndef BOOL
#define BOOL char
#endif
//#ifndef LWORD
//#define LWORD unsigned long
//#endif

#define outnic(addr, data) XIo_Out16(NIC_BASE+addr, data)
#define innic(addr) (XIo_In16(NIC_BASE+addr))

#define PROMISC  0              /* Set non-zero to accept all packets */

/* Ethernet definitions.. */
#define MINFRAME  60
#define MINFRAMEC 64
#define CRCLEN    4
#define MAXFRAME  1514
#define MAXFRAMEC 1518

/* NE2000 definitions */
#define NIC_BASE (0x00A00400) // Base I/O address of the NIC card
#define DATAPORT (0x10*2)
#define NE_RESET (0x1f*2)

/* 8390 Network Interface Controller (NIC) page0 register offsets */
#define CMDR    (0x00*2)            /* command register for read & write */
#define PSTART  (0x01*2)            /* page start register for write */
#define PSTOP   (0x02*2)            /* page stop register for write */
#define BNRY    (0x03*2)            /* boundary reg for rd and wr */
#define TPSR    (0x04*2)            /* tx start page start reg for wr */
#define TBCR0   (0x05*2)            /* tx byte count 0 reg for wr */
#define TBCR1   (0x06*2)            /* tx byte count 1 reg for wr */
#define ISR     (0x07*2)            /* interrupt status reg for rd and wr */
#define RSAR0   (0x08*2)            /* low byte of remote start addr */
#define RSAR1   (0x09*2)            /* hi byte of remote start addr */
#define RBCR0   (0x0A*2)            /* remote byte count reg 0 for wr */
#define RBCR1   (0x0B*2)            /* remote byte count reg 1 for wr */
#define RCR     (0x0C*2)            /* rx configuration reg for wr */
#define TCR     (0x0D*2)            /* tx configuration reg for wr */
#define DCR     (0x0E*2)            /* data configuration reg for wr */
#define IMR     (0x0F*2)            /* interrupt mask reg for wr */

/* NIC page 1 register offsets */
#define PAR0    (0x01*2)            /* physical addr reg 0 for rd and wr */
#define CURRP   (0x07*2)            /* current page reg for rd and wr */
#define MAR0    (0x08*2)            /* multicast addr reg 0 for rd and WR */

/* NIC page 3 register offsets */
//#define RTL9346CR 0x01          /* RTL 9346 command reg */
//#define RTL3    0x06            /* RTL config reg 3 */

/* NIC RAM definitions */
//#define RAMPAGES 0x20           /* Total number of 256-byte RAM pages */
#define TXSTART  0x41           /* Tx buffer start page */
#define TXPAGES  8              /* Pages for Tx buffer */
#define RXSTART  (TXSTART+TXPAGES)      /* Rx buffer start page */
#define RXSTOP   0x80   /* Last Rx buffer page */
#define DCRVAL   0x00           /* Value for data config reg for word-wide tx */

/* IP protocol definitions */

#define PCOL_ARP    0x0806  // Protocol type: ARP
#define PCOL_IP     0x0800  //                IP

#define ARPREQ     0x0001   // ARP request & response IDs
#define ARPRESP    0x0002

#define PICMP       1       // IP protocol values: ICMP
#define PTCP        6       //                     TCP
#define PUDP        17      //                     UDP

#define ETHHDR_LEN  6       // Ethernet frame header length
#define IPHDR_LEN   20      // IP, TCP and ICMP header lengths
#define TCPHDR_LEN  20
#define ICMPHDR_LEN 4       // (only include type, code & csum in ICMP hdr)
#define TCPOPT_LEN  4       // Length of TCP MSS option
#define TCPSYN_LEN  24      // TCP header length including MSS option
#define MAXPING_LEN 212     // Max length of Ping data

#define TFIN        0x01    // Option flags: no more data
#define TSYN        0x02    //           sync sequence nums
#define TRST        0x04    //           reset connection
#define TPUSH       0x08    //           push buffered data
#define TACK        0x10    //           acknowledgement
#define TURGE       0x20    //           urgent

#define TCP_MSS     1460    // Max Segment Size for TCP

#define ECHOPORT    7       // TCP Port numbers: echo
#define DAYPORT     13      //                   daytime
#define HTTPORT     80      //                   HTTP

/* end IP definitions */


#define MACLEN   6
BYTE myeth[MACLEN] = {0x00, 0x04, 0xa3, 0, 0, 0};
int next_page;
BYTE curr_rx_page;


typedef struct {                // NIC hardware packet header
    BYTE stat;                  //     Error status
    BYTE next;                  //     Pointer to next block
    WORD len;                   //     Length of this frame incl. CRC
} NICHEADER;

typedef struct {                // Ethernet frame header
    BYTE dest[MACLEN];          //     Dest & srce MAC addresses
    BYTE srce[MACLEN];
    WORD pcol;                  //     Protocol
} ETHERHEADER;

typedef struct {                // NIC and Ethernet headers combined
    NICHEADER nic;
    ETHERHEADER eth;
} NICETHERHEADER;

NICETHERHEADER nicin;           // Buffer for incoming NIC & Ether hdrs

/* Globals from driver.h */
BOOL checkflag;             // Checksum flag & byte values
BYTE checkhi, checklo;

BYTE ungot_byte;
BOOL ungot;
/* end driver.h globals */

WORD rxin, rxout;               // Rx buffer counts (NIC RAM is buffer)
BOOL atend;

#define TXBUFFLEN 64
BYTE txbuff[TXBUFFLEN];         // Tx buffer
int txin, txout;

/* Prototypes */
//void setnic_addr(WORD addr);
WORD init_ether(void);
WORD diagnostics(void);
WORD delay_ms(WORD mult);
void swapw(WORD *val);
void putnic_data(BYTE *data, WORD len);//
void setnic_addr(WORD addr);
void put_ether(void *data, WORD dlen); //
void xmit_ether(WORD dlen);
WORD get_ether(void);
void getnic_data(BYTE *data, int len);//
BYTE nicwrap(WORD page);
void check_byte(BYTE b);
BYTE getch_net(void);
BYTE getnic_byte(void);
BOOL match_byte(BYTE b);
BOOL get_word(WORD *w);
BOOL match_word(WORD w);
BOOL skip_lword(void);
BOOL skip_word(void);
BOOL get_lword(BYTE *lw);
BOOL match_lword(BYTE *lw);
void putch_net(BYTE b);
void put_word(WORD w);
void put_byte(BYTE b);
void put_data(BYTE *data, WORD len);//
void put_lword(BYTE *lw);
void discard_data(void);
void check_lword(BYTE *lw);
void put_ip(void);
BOOL skip_byte(void);
void copy_rx_tx(BYTE dest, BYTE srce, BYTE len);
void putnic_byte(BYTE b);
BOOL arp_recv(void);

/* from p16_ip.h from tcp book lib */
BYTE ipcol;                 // IP protocol byte
BYTE myip[4];               // My IP adress
BYTE locip[4], remip[4];    // Local & remote IP addresses

int locport, remport;      // ..and TCP port numbers
BYTE rseq[4], rack[4];      // TCP sequence & acknowledge values
int concount;              // Connection count (for high word of my seq num)
BYTE rflags, tflags;        // Rx and Tx flags
int rpdlen, tpdlen;        // Length of user data in Rx, Tx buffer
signed long iplen;          // Incoming/outgoing IP length word
BYTE d_checkhi, d_checklo;  // Checksum value for data
/* end tcp book lib */

/* Initialize card given driver type and base addr.
** Return driver type, 0 if error */
WORD init_ether(void)
{
    WORD ok=0, i;
    BYTE b;

    //reset_ether();
    //delay_ms(2);
    //NIC_RESET = 0;
    delay_ms(2);
    outnic(NE_RESET, innic(NE_RESET));  /* Do reset */
    delay_ms(2);
    if ((innic(ISR) & 0x80) == 0)       /* Report if failed */
    {
        print("NIC init error\r\n");
    }
    else
    {
        print("Ethernet card init successful!\r\n");
        outnic(CMDR, 0x21);                 /* Stop, DMA abort, page 0 */
        delay_ms(2);                        /* ..wait to take effect */
        outnic(DCR, DCRVAL);
        outnic(RBCR0, 0x00);                   /* Clear remote byte count */
        outnic(RBCR1, 0x00);
        outnic(RCR, 0x20);                  /* Rx monitor mode */
        outnic(TCR, 0x02);                  /* Tx internal loopback */
        outnic(TPSR, TXSTART);              /* Set Tx start page */
        outnic(PSTART, RXSTART);            /* Set Rx start, stop, boundary */
        outnic(PSTOP, RXSTOP);
        outnic(BNRY, (BYTE)(RXSTOP-1));
        outnic(ISR, 0xff);                  /* Clear interrupt flags */
        outnic(IMR, 0x00);                     /* Mask all interrupts */
        outnic(CMDR, 0x61);                 /* Stop, DMA abort, page 1 */
        delay_ms(2);
        for (i=0; i<6; i++)                 /* Set Phys addr */
            outnic(PAR0+i, myeth[i]);
        for (i=0; i<8; i++)                 /* Multicast accept-all */
            outnic(MAR0+i, 0xff);
        outnic(CURRP, RXSTART+1);           /* Set current Rx page */
        next_page = RXSTART + 1;
// Set LED 0 to be a 'link' LED, not 'collision' LED
// It would be nice if the following code worked, but the upper bits of the
// RTL config3 register are obstinately read-only, so it doesn't!
//        outnic(CMDR, 0xe0);                 /* DMA abort, page 3 */
//        outnic(RTL9346CR, 0xc0);            /* Write-enable config regs */
//        outnic(RTL3, 0x10);                 /* Enable 'link' LED */
//        outnic(RTL9346CR, 0x00);            /* Write-protect config regs */
        outnic(CMDR, 0x20);                 /* DMA abort, page 0 */
	//#if PROMISC
        //outnic(RCR, 0x14);                  /* Accept broadcasts and all packets!*/
	//#else
        outnic(RCR, 0x04);                  /* Accept broadcasts */
	//#endif
        outnic(TCR, 0x00);                     /* Normal Tx operation */
        outnic(ISR, 0xff);                  /* Clear interrupt flags */
        outnic(CMDR, 0x22);                 /* Start NIC */
        ok = 1;
    }
    return(ok);
}



WORD diagnostics(void) {
  //  outnic(CMDR, 0x21);                // stop AX88796
  delay_ms(2);

  outnic(CMDR,0x61);
  outnic(PSTART, 0x4E);

  outnic(CMDR,0x21);
  outnic(PSTOP, 0x3E);

  outnic(CMDR,0x61);
  if(innic(PSTART) != 0x4e)
    return 1;

  outnic(CMDR,0x21);                 // switch to page 0
  if(innic(PSTOP) != 0x3e)
    return 2;

  if(innic(0x16*2) != 0x15)
    return 3;

  if(innic(0x12*2) != 0x0c)
    return 4;

  if(innic(0x13*2) != 0x12)
    return 5;

  return 0;
}

WORD delay_ms(WORD mult){
  WORD i;
  WORD delay = 5000000*mult;
  for(i=0; i<delay; i++);
  return 1;
}

/* Swap the bytes in a word */
void swapw(WORD *val)
{
  WORD temp_word;
  temp_word = *val;
    *val = (temp_word >> 8) | ((temp_word << 8) & 0xff00);
}

void putnic_data(BYTE *data, WORD len)
{
    len += len & 1;                     /* Round length up to an even value */
    outnic(ISR, 0x40);                  /* Clear remote DMA interrupt flag */
    outnic(RBCR0, len);                 /* Byte count */
    outnic(RBCR1, 0);
    outnic(CMDR, 0x12);                 /* Start, DMA remote write */
    while (len--)                        /* O/P bytes */
        outnic(DATAPORT, *data++);
    len = 255;                          /* Done: must ensure DMA complete */
    while (len && (innic(ISR)&0x40)==0)
        len--;
}

/* Set the 'remote DMA' address in the NIC's RAM to be accessed */
void setnic_addr(WORD addr)
{
    outnic(ISR, 0x40);                  /* Clear remote DMA interrupt flag */
    outnic(RSAR0, addr&0xff);           /* Data addr */
    outnic(RSAR1, addr>>8);
}

/* Send Ethernet packet given payload len */
void put_ether(void *data, WORD dlen)
{
    outnic(ISR, 0x0a);              /* Clear interrupt flags */
    setnic_addr(TXSTART<<8);
    putnic_data(nicin.eth.srce, MACLEN);
    putnic_data(myeth, MACLEN);
    swapw(&nicin.eth.pcol);
    putnic_data(&nicin.eth.pcol, 2);
    putnic_data(data, dlen);
}

/* Transmit the Ethernet frame */
void xmit_ether(WORD dlen)
{
    dlen += MACLEN+MACLEN+2;
    if (dlen < MINFRAME)
        dlen = MINFRAME;            /* Constrain length */
    outnic(TBCR0, dlen);            /* Set Tx length regs */
    outnic(TBCR1, dlen >> 8);
    outnic(CMDR, 0x24);             /* Transmit the packet */
}

/* Get packet into buffer, return length (excl CRC), or 0 if none available */

WORD get_ether(void)
{
    WORD len=0, curr;
    BYTE bound;
    WORD oset;
    WORD i;

    if (innic(ISR) & 0x10)              /* If Rx overrun.. */
    {
      //printf(" NIC Rx overrun ");
        init_ether();                   /* ..reset controller (drastic!) */
    }
    outnic(CMDR, 0x60);                 /* DMA abort, page 1 */
    curr = innic(CURRP);                /* Get current page */
    outnic(CMDR, 0x20);                 /* DMA abort, page 0 */
    if (curr != next_page)              /* If Rx packet.. */
    {
        //putnum(curr);
        //print(" ");

        //putnum(next_page);
        //print("\r\n");

        curr_rx_page = next_page;
        setnic_addr((WORD)next_page<<8);

	getnic_data((BYTE *)&nicin, sizeof(nicin));

	//print("######nicin: ");
	//putnum(nicin.nic.stat);
	//print(" ");
	//putnum(nicin.nic.next);
	//print(" ");
	//putnum(nicin.nic.len);
	//print(" ");
	//putnum(nicin.eth.dest);
	//print(" ");
	//putnum(nicin.eth.srce);
	//print(" ");
	//putnum(nicin.eth.pcol);
	//print("\r\n");

            swapw(&nicin.nic.len);//
            //swapw(&nicin.eth.pcol);//



        len = nicin.nic.len;            /* Take length from stored header */

        if ((nicin.nic.stat&1) && len>=MINFRAMEC && len<=MAXFRAMEC)
        {                               /* If hdr is OK, get packet */
            len -= MACLEN+MACLEN+2+CRCLEN;
            rxin = len;
            //swapw(&nicin.eth.pcol);
	    //#if DEBUG
            //printf(" Rx%lu>", len);
	    //#endif
        }
        else                            /* If not, no packet data */
        {
	  //if((nicin.nic.stat&1))
	    //{
	      //print("stat ok\r\n");
	    //}
	  //if(len>=MINFRAMEC)
	    //{
	      //print("min ok\r\n");
	    //}
	  //if(len<=MAXFRAMEC)
	    //{
	      //print("max ok\r\n");
	    //}

            len = 0;
            print("NIC pkt err \r\n");
        }

	                                /* Update next packet ptr */
        if (nicin.nic.next>=RXSTART && nicin.nic.next<RXSTOP)
            next_page = nicin.nic.next;
        else                            /* If invalid, use prev+1 */
        {
	  //printf(" NIC ptr err ");
            next_page = nicwrap(next_page + 1);
            len = 0;
        }                               /* Update boundary register */
        bound = nicwrap(next_page - 1);
        outnic(BNRY, bound);
    }
    return(len);                        /* Return length excl. CRC */
}

/* Get data from NIC's RAM into the given buffer */
void getnic_data(BYTE *data, int len)
{
    BYTE b;

    outnic(ISR, 0x40);                  /* Clear remote DMA interrupt flag */
    outnic(RBCR0, len&0xff);                 /* Byte count */
    outnic(RBCR1, 0);
    outnic(CMDR, 0x0a);                 /* Start, DMA remote read */
    while (len--)                       /* Get bytes */
    {
        b = innic(DATAPORT);
	//print("** ");
        *data++ = b;
	//putnum(b);
	//print("\r\n");
    }
    //print("\r\n");
}

/* Wrap an NIC Rx page number */
BYTE nicwrap(WORD page)
{
   if (page >= RXSTOP)
       page += RXSTART - RXSTOP;
   else if (page < RXSTART)
       page += RXSTOP - RXSTART;
   return(page);
}


/*-----------------START IP PROTOCOLS---------------------*/

/* Handle an ARP message */
BOOL arp_recv(void)
{
    BOOL ret = 0;

   //if(nic

    //DEBUG_PUTC('a');
    if (match_byte(0x00) && match_byte(0x01) &&     // Hardware type
        match_byte(0x08) && match_byte(0x00) &&     // ARP protocol
        match_byte(6) &&  match_byte(4) &&          // Hardware & IP lengths
        match_word(ARPREQ) &&                       // ARP request
        skip_lword() && skip_word() &&              // Sender's MAC addr
        get_lword(remip) &&                       // Sender's IP addr
        skip_lword() && skip_word() &&              // Null MAC addr
        match_lword(myip))                        // Target IP addr (me?)
    {
      //DEBUG_PUTC('>');
      //print("0000000");
        ret = 1;
        txin = 0;
	put_word(0x0001);                           // Hardware type
        put_word(0x0800);                           // ARP protocol
        put_byte(6);                                // Hardware & IP lengths
        put_byte(4);
        put_word(ARPRESP);                          // ARP response
        put_data(myeth, MACLEN);                    // My MAC addr
        put_lword(myip);                          // My IP addr
        put_data(nicin.eth.srce, MACLEN);           // Remote MAC addr
        put_lword(remip);                         // Remote IP addr
        put_ether(txbuff, txin);                    // Send to NIC
        xmit_ether(txin);                           // Transmit
        //DEBUG_PUTC('A');
    }
    else {
      print("out");
      discard_data();
      }
    return(ret);
}

/* Add byte to checksum value */
void check_byte(BYTE b)
{
    if (checkflag)
    {
        if ((checklo = b+checklo) < b)
        {
            if (++checkhi == 0)
                checklo++;
        }
    }
    else
    {
        if ((checkhi = b+checkhi) < b)
        {
            if (++checklo == 0)
                checkhi++;
        }
    }
    checkflag = !checkflag;
}

/* Get a byte from network buffer; if end, set flag */
BYTE getch_net(void)
{
    BYTE b=0;

    atend = rxout >= rxin;
//print("!1");
//putnum(!atend);
//print("2!");
    if (!atend)
    {
    	//print("yea");
        b = getnic_byte();
	putnum(b);
	print("----\r\n");
        rxout++;
        check_byte(b);
    }
    return(b);
}

/* Return a byte from the NIC RAM */
BYTE getnic_byte(void)
{
    BYTE b;

    outnic(RBCR0, 1);                   /* Byte count */
    outnic(RBCR1, 0);
    outnic(CMDR, 0x0a);                 /* Start, DMA remote read */
    b = innic(DATAPORT);
    return(b);
}

/* Match an incoming byte value, return 0 not matched, or end of message */
BOOL match_byte(BYTE b)
{

    //return(b==getch_net() && !atend);
    BYTE a;
    a = getch_net();
    print("byte: ");
    putnum(a);
    print("\r\n");
    print("correct: ");
    putnum(b);
    print("\r\n");
    return(b==a && !atend);
}

/* Get an incoming word value, return 0 if end of message */
BOOL get_word(WORD *w)
{
    BYTE hi, lo;

    hi = getch_net();
    lo = getch_net();
    w = ((WORD)hi<<8) | (WORD)lo;
    return(!atend);
}

/* Match an incoming byte value, return 0 not matched, or end of message */
BOOL match_word(WORD w)
{
    WORD inw;

    return(get_word(inw) && inw==w);
}

/* Skip an incoming lword value, return 0 if end of message */
BOOL skip_lword(void)
{
    getch_net();
    getch_net();
    getch_net();
    getch_net();
    return(!atend);
}

/* Skip an incoming word value, return 0 if end of message */
BOOL skip_word(void)
{
    getch_net();
    getch_net();
    return(!atend);
}

/* Get an incoming lword value, return 0 if end of message */
//BOOL get_lword(LWORD *lw)
BOOL get_lword(BYTE *lw)
{
  lw[3] = getch_net();//
  lw[2] = getch_net();//
  lw[1] = getch_net();//
  lw[0] = getch_net();//
  //lw.b[3] = getch_net();
  //lw.b[2] = getch_net();
  //lw.b[1] = getch_net();
  //lw.b[0] = getch_net();
  return(!atend);
}

/* Match longword */
//BOOL match_lword(LWORD *lw)
BOOL match_lword(BYTE *lw)
{
  //return (match_byte(lw.b[3]) && match_byte(lw.b[2]) &&
  //    match_byte(lw.b[1]) && match_byte(lw.b[0]));
  return (match_byte(lw[3]) && match_byte(lw[2]) &&
        match_byte(lw[1]) && match_byte(lw[0]));
}

/* Put a byte into the network buffer */
void putch_net(BYTE b)
{
    if (txin < TXBUFFLEN)
        txbuff[txin++] = b;
    check_byte(b);
}

/* Send a word out to the SLIP link, then add to checksum */
void put_word(WORD w)
{
    putch_net(w >> 8);
    putch_net(w);
}

/* Send a byte to the network buffer */
void put_byte(BYTE b)
{
    putch_net(b);
}

void put_data(BYTE *data, WORD len)//
{
    while (len--)
        putch_net(*data++);
}

//void put_lword(LWORD *lw)
void put_lword(BYTE *lw)
{
  //putch_net(lw.b[3]);
  //putch_net(lw.b[2]);
  //putch_net(lw.b[1]);
  //putch_net(lw.b[0]);
    putch_net(lw[3]);
    putch_net(lw[2]);
    putch_net(lw[1]);
    putch_net(lw[0]);
}

/* Discard incoming data */
void discard_data(void)
{
    while (!atend)
        getch_net();
}

/* Respond to an ICMP message (e.g. ping) */
BOOL icmp_recv(void)
{
    BOOL ret=0;
    WORD csum;

    //DEBUG_PUTC('c');
    rpdlen = 0;
    if (match_byte(8) && match_byte(0) && get_word(csum))
    {
        while (skip_byte())                 // Check data
            rpdlen++;
        ret = (checkhi==0xff) && (checklo==0xff);
        if (ret && rpdlen<=MAXPING_LEN) 
        {                                   // If OK and not bcast..
	  //DEBUG_PUTC('>');
            checkhi = checklo = 0;          // Clear checksum
            put_ip();                       // IP header
            put_word(0);                    // ICMP type and code
            csum += 0x0800;                 // Adjust checksum for resp
            if (csum < 0x0800)              // ..including hi-lo carry
                csum++;
            put_word(csum);                 // ICMP checksum
            put_ether(txbuff, txin);        // Send ICMP response
            copy_rx_tx(txin, IPHDR_LEN+ICMPHDR_LEN, rpdlen);
            xmit_ether(IPHDR_LEN+ICMPHDR_LEN+rpdlen);
            //DEBUG_PUTC('I');
        }
    }
    return(ret);
}

//void check_lword(LWORD *lw)
void check_lword(BYTE *lw)
{
  //check_byte(lw.b[3]);
  //check_byte(lw.b[2]);
  //check_byte(lw.b[1]);
  //check_byte(lw.b[0]);
    check_byte(lw[3]);
    check_byte(lw[2]);
    check_byte(lw[1]);
    check_byte(lw[0]);
}

/* Send out an IP datagram header, given data length */
void put_ip(void)
{
    static BYTE id=0;

    txin = 0;
    checkhi = checklo = 0;          // Clear checksum
    checkflag = 0;
    put_byte(0x45);                 // Version & hdr len */
    put_byte(0);                    // Service
    put_word(iplen);
    put_byte(0);                    // Ident word
    put_byte(++id);
    put_word(0);                    // Flags & fragment offset
    put_byte(100);                  // Time To Live
    put_byte(ipcol);                // Protocol
    check_lword(myip);            // Include addresses in checksum
    check_lword(remip);
    put_byte(~checkhi);             // Checksum
    put_byte(~checklo);
    put_lword(myip);              // Source & destination IP addrs
    put_lword(remip);
}

/* Skip an incoming byte value, return 0 if end of message */
BOOL skip_byte(void)
{
    getch_net();
    return(!atend);
}

/* Copy a block from NIC Rx to Tx buffers (not crossing page boundaries) */
void copy_rx_tx(BYTE dest, BYTE srce, BYTE len)
{       
    BYTE b;
    
    outnic(ISR, 0x40);                  /* Clear remote DMA interrupt flag */
    dest += sizeof(ETHERHEADER);
    srce += sizeof(NICETHERHEADER);
    while (len--)
    {
        outnic(RSAR0, srce);
        outnic(RSAR1, curr_rx_page);
        b = getnic_byte();
        outnic(RSAR0, dest);
        outnic(RSAR1, TXSTART);
        putnic_byte(b);
        srce++;
        dest++;
    }
}

/* Put the given byte into the NIC's RAM */
void putnic_byte(BYTE b)
{
    outnic(RBCR0, 1);                   /* Byte count */
    outnic(RBCR1, 0);
    outnic(CMDR, 0x12);                 /* Start, DMA remote write */
    outnic(DATAPORT, b);
}
/* EOF */
